# Utility functions for the package
# Copyright (C) 2019  Sidney Barthe, Robin Scheibler, Ivan Dokmanic, Eric Bezzam
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import numpy as np
from scipy import signal
from scipy.io import wavfile

from .parameters import constants, eps
from .sync import correlate


def create_noisy_signal(signal_fp, snr, noise_fp=None, offset=None):
    """
    Create a noisy signal of a specified SNR.
    Parameters
    ----------
    signal_fp : string
        File path to clean input.
    snr : float
        SNR in dB.
    noise_fp : string
        File path to noise. Default is to use randomly generated white noise.
    offset : float
        Offset in seconds before starting the signal.

    Returns
    -------
    numpy array
        Noisy signal with specified SNR, between [-1,1] and zero mean.
    numpy array
        Clean signal, untouched from WAV file.
    numpy array
        Added noise such that specified SNR is met.
    int
        Sampling rate in Hz.

    """
    fs, clean_signal = wavfile.read(signal_fp)
    clean_signal = to_float32(clean_signal)

    if offset is not None:
        offset_samp = int(offset * fs)
    else:
        offset_samp = 0
    output_len = len(clean_signal) + offset_samp

    if noise_fp is not None:
        fs_n, noise = wavfile.read(noise_fp)
        noise = to_float32(noise)
        if fs_n != fs:
            raise ValueError(
                "Signal and noise WAV files should have same " "sampling rate."
            )
        # truncate to same length
        if len(noise) < output_len:
            raise ValueError(
                "Length of signal file should be longer than " "noise file."
            )
        noise = noise[:output_len]
    else:
        if len(clean_signal.shape) > 1:  # multichannel
            noise = np.random.randn(output_len, clean_signal.shape[1]).astype(
                np.float32
            )
        else:
            noise = np.random.randn(output_len).astype(np.float32)
        noise = normalize(noise)

    # weight noise according to desired SNR
    signal_level = rms(clean_signal)
    noise_level = rms(noise[offset_samp:])
    noise_fact = signal_level / noise_level * 10 ** (-snr / 20)
    noise_weighted = noise * noise_fact

    # add signal and noise
    noisy_signal = clean_signal + noise_weighted

    # ensure between [-1, 1]
    norm_fact = np.abs(noisy_signal).max()
    clean_signal /= norm_fact
    noise_weighted /= norm_fact
    noisy_signal /= norm_fact

    # remove any offset
    noisy_signal -= noisy_signal.mean()

    return noisy_signal, clean_signal, noise_weighted, fs


def rms(data):
    """
    Compute root mean square of input.

    Parameters
    ----------
    data : numpy array
        Real signal in time domain.

    Returns
    -------
    float
        Root mean square.
    """
    return np.sqrt(np.mean(data * data))


def to_float32(data):
    """
    Cast data (typically from WAV) to float32.

    Parameters
    ----------
    data : numpy array
        Real signal in time domain, typically obtained from WAV file.

    Returns
    -------
    numpy array
        `data` as float32.
    """

    if np.issubdtype(data.dtype, np.integer):
        max_val = abs(np.iinfo(data.dtype).min)
        return data.astype(np.float32) / max_val
    else:
        return data.astype(np.float32)


def to_16b(signal):
    """
    converts float 32 bit signal (-1 to 1) to a signed 16 bits representation
    No clipping in performed, you are responsible to ensure signal is within
    the correct interval.
    """
    return ((2 ** 15 - 1) * signal).astype(np.int16)


def clip(signal, high, low):
    """Clip a signal from above at high and from below at low."""
    s = signal.copy()

    s[np.where(s > high)] = high
    s[np.where(s < low)] = low

    return s


def normalize(signal, bits=None):
    """
    normalize to be in a given range. The default is to normalize the maximum
    amplitude to be one. An optional argument allows to normalize the signal
    to be within the range of a given signed integer representation of bits.
    """

    s = signal.copy()

    s /= np.abs(s).max()

    # if one wants to scale for bits allocated
    if bits is not None:
        s *= 2 ** (bits - 1) - 1
        s = clip(s, 2 ** (bits - 1) - 1, -(2 ** (bits - 1)))

    return s


def normalize_pwr(sig1, sig2):
    """Normalize sig1 to have the same power as sig2."""

    # average power per sample
    p1 = np.mean(sig1 ** 2)
    p2 = np.mean(sig2 ** 2)

    # normalize
    return sig1.copy() * np.sqrt(p2 / p1)


def highpass(signal, Fs, fc=None, plot=False):
    """ Filter out the really low frequencies, default is below 50Hz """

    if fc is None:
        fc = constants.get("fc_hp")

    # have some predefined parameters
    rp = 5  # minimum ripple in dB in pass-band
    rs = 60  # minimum attenuation in dB in stop-band
    n = 4  # order of the filter
    type = "butter"

    # normalized cut-off frequency
    wc = 2.0 * fc / Fs

    # design the filter
    from scipy.signal import iirfilter, lfilter, freqz

    b, a = iirfilter(n, Wn=wc, rp=rp, rs=rs, btype="highpass", ftype=type)

    # plot frequency response of filter if requested
    if plot:
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings

            warnings.warn("Matplotlib is required for plotting")
            return

        w, h = freqz(b, a)

        plt.figure()
        plt.title("Digital filter frequency response")
        plt.plot(w, 20 * np.log10(np.abs(h)))
        plt.title("Digital filter frequency response")
        plt.ylabel("Amplitude Response [dB]")
        plt.xlabel("Frequency (rad/sample)")
        plt.grid()

    # apply the filter
    signal = lfilter(b, a, signal.copy())

    return signal


def time_dB(signal, Fs, bits=16):
    """
    Compute the signed dB amplitude of the oscillating signal
    normalized wrt the number of bits used for the signal.
    """

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        import warnings

        warnings.warn("Matplotlib is required for plotting")
        return

    # min dB (least significant bit in dB)
    lsb = -20 * np.log10(2.0) * (bits - 1)

    # magnitude in dB (clipped)
    pos = clip(signal, 2.0 ** (bits - 1) - 1, 1.0) / 2.0 ** (bits - 1)
    neg = -clip(signal, -1.0, -(2.0 ** (bits - 1))) / 2.0 ** (bits - 1)

    mag_pos = np.zeros(signal.shape)
    Ip = np.where(pos > 0)
    mag_pos[Ip] = 20 * np.log10(pos[Ip]) + lsb + 1

    mag_neg = np.zeros(signal.shape)
    In = np.where(neg > 0)
    mag_neg[In] = 20 * np.log10(neg[In]) + lsb + 1

    plt.plot(np.arange(len(signal)) / float(Fs), mag_pos - mag_neg)
    plt.xlabel("Time [s]")
    plt.ylabel("Amplitude [dB]")
    plt.axis("tight")
    plt.ylim(lsb - 1, -lsb + 1)

    # draw ticks corresponding to decibels
    div = 20
    n = int(-lsb / div) + 1
    yticks = np.zeros(2 * n)
    yticks[:n] = lsb - 1 + np.arange(0, n * div, div)
    yticks[n:] = -lsb + 1 - np.arange((n - 1) * div, -1, -div)
    yticklabels = np.zeros(2 * n)
    yticklabels = range(0, -n * div, -div) + range(-(n - 1) * div, 1, div)
    plt.setp(plt.gca(), "yticks", yticks)
    plt.setp(plt.gca(), "yticklabels", yticklabels)

    plt.setp(plt.getp(plt.gca(), "ygridlines"), "ls", "--")


def spectrum(signal, Fs, N):

    from .stft import stft, spectroplot
    from .windows import hann

    F = stft(signal, N, N / 2, win=hann(N))
    spectroplot(F.T, N, N / 2, Fs)


def dB(signal, power=False):
    if power is True:
        return 10 * np.log10(np.abs(signal))
    else:
        return 20 * np.log10(np.abs(signal))


def compare_plot(
    signal1,
    signal2,
    Fs,
    fft_size=512,
    norm=False,
    equal=False,
    title1=None,
    title2=None,
):

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        import warnings

        warnings.warn("Matplotlib is required for plotting")
        return

    td_amp = np.maximum(np.abs(signal1).max(), np.abs(signal2).max())

    if norm:
        if equal:
            signal1 /= np.abs(signal1).max()
            signal2 /= np.abs(signal2).max()
        else:
            signal1 /= td_amp
            signal2 /= td_amp
        td_amp = 1.0

    plt.subplot(2, 2, 1)
    plt.plot(np.arange(len(signal1)) / float(Fs), signal1)
    plt.axis("tight")
    plt.ylim(-td_amp, td_amp)
    if title1 is not None:
        plt.title(title1)

    plt.subplot(2, 2, 2)
    plt.plot(np.arange(len(signal2)) / float(Fs), signal2)
    plt.axis("tight")
    plt.ylim(-td_amp, td_amp)
    if title2 is not None:
        plt.title(title2)

    from .stft import stft, spectroplot
    from .windows import hann

    F1 = stft.stft(signal1, fft_size, fft_size / 2, win=windows.hann(fft_size))
    F2 = stft.stft(signal2, fft_size, fft_size / 2, win=windows.hann(fft_size))

    # try a fancy way to set the scale to avoid having the spectrum
    # dominated by a few outliers
    p_min = 1
    p_max = 99.5
    all_vals = np.concatenate((dB(F1 + eps), dB(F2 + eps))).flatten()
    vmin, vmax = np.percentile(all_vals, [p_min, p_max])

    cmap = "jet"
    interpolation = "sinc"

    plt.subplot(2, 2, 3)
    stft.spectroplot(
        F1.T,
        fft_size,
        fft_size / 2,
        Fs,
        vmin=vmin,
        vmax=vmax,
        cmap=plt.get_cmap(cmap),
        interpolation=interpolation,
    )

    plt.subplot(2, 2, 4)
    stft.spectroplot(
        F2.T,
        fft_size,
        fft_size / 2,
        Fs,
        vmin=vmin,
        vmax=vmax,
        cmap=plt.get_cmap(cmap),
        interpolation=interpolation,
    )


def real_spectrum(signal, axis=-1, **kwargs):

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        import warnings

        warnings.warn("Matplotlib is required for plotting")
        return

    S = np.fft.rfft(signal, axis=axis)
    f = np.arange(S.shape[axis]) / float(2 * S.shape[axis])

    plt.subplot(2, 1, 1)
    P = dB(S)
    plt.plot(f, P, **kwargs)

    plt.subplot(2, 1, 2)
    phi = np.unwrap(np.angle(S))
    plt.plot(f, phi, **kwargs)


def convmtx(x, n):
    """
    Create a convolution matrix H for the vector x of size len(x) times n.
    Then, the result of np.dot(H,v) where v is a vector of length n is the same
    as np.convolve(x, v).
    """

    import scipy as s

    c = np.concatenate((x, np.zeros(n - 1)))
    r = np.zeros(n)

    return s.linalg.toeplitz(c, r)


def prony(x, p, q):
    """
    Prony's Method from Monson H. Hayes' Statistical Signal Processing, p. 154

    Parameters
    ----------

    x: 
        signal to model
    p: 
        order of denominator
    q: 
        order of numerator

    Returns
    -------

    a: 
        numerator coefficients
    b: 
        denominator coefficients
    err: the squared error of approximation
    """

    nx = x.shape[0]

    if p + q >= nx:
        raise NameError("Model order too large")

    X = convmtx(x, p + 1)

    Xq = X[q : nx + p - 1, 0:p]

    a = np.concatenate(
        (np.ones(1), -np.linalg.lstsq(Xq, X[q + 1 : nx + p, 0], rcond=None)[0])
    )
    b = np.dot(X[0 : q + 1, 0 : p + 1], a)

    err = np.inner(np.conj(x[q + 1 : nx]), np.dot(X[q + 1 : nx, : p + 1], a))

    return a, b, err


def shanks(x, p, q):
    """
    Shank's Method from Monson H. Hayes' Statistical Signal Processing, p. 154

    Parameters
    ----------
    x: 
        signal to model
    p: 
        order of denominator
    q: 
        order of numerator

    Returns
    -------
    a: 
        numerator coefficients
    b: 
        denominator coefficients
    err: 
        the squared error of approximation
    """

    from scipy import signal

    nx = x.shape[0]

    if p + q >= nx:
        raise NameError("Model order too large")

    a = prony(x, p, q)[0]

    u = np.zeros(nx)
    u[0] = 1.0

    g = signal.lfilter(np.ones(1), a, u)

    G = convmtx(g, q + 1)
    b = np.linalg.lstsq(G[:nx, :], x, rcond=None)[0]
    err = np.inner(np.conj(x), x) - np.inner(np.conj(x), np.dot(G[:nx, : q + 1], b))

    return a, b, err


def low_pass_dirac(t0, alpha, Fs, N):
    """
    Creates a vector containing a lowpass Dirac of duration T sampled at Fs
    with delay t0 and attenuation alpha.

    If t0 and alpha are 2D column vectors of the same size, then the function
    returns a matrix with each line corresponding to pair of t0/alpha values.
    """

    return alpha * np.sinc(np.arange(N) - Fs * t0)


def fractional_delay(t0):
    """
    Creates a fractional delay filter using a windowed sinc function.
    The length of the filter is fixed by the module wide constant
    `frac_delay_length` (default 81).

    Parameters
    ----------
    t0: float
        The delay in fraction of sample. Typically between -1 and 1.

    Returns
    -------
    numpy array
        A fractional delay filter with specified delay.
    """

    N = constants.get("frac_delay_length")

    return np.hanning(N) * np.sinc(np.arange(N) - (N - 1) / 2 - t0)


def fractional_delay_filter_bank(delays):
    """
    Creates a fractional delay filter bank of windowed sinc filters

    Parameters
    ----------
    delays: 1d narray
        The delays corresponding to each filter in fractional samples
        
    Returns
    -------
    numpy array
        An ndarray where the ith row contains the fractional delay filter
        corresponding to the ith delay. The number of columns of the matrix
        is proportional to the maximum delay.
    """

    delays = np.array(delays)

    # subtract the minimum delay, so that all delays are positive
    delays -= delays.min()

    # constants and lengths
    N = delays.shape[0]
    L = constants.get("frac_delay_length")
    filter_length = L + int(np.ceil(delays).max())

    # allocate a flat array for the filter bank that we'll reshape at the end
    bank_flat = np.zeros(N * filter_length)

    # separate delays in integer and fractional parts
    di = np.floor(delays).astype(np.int)
    df = delays - di

    # broadcasting tricks to compute at once all the locations
    # and sinc times that must be computed
    T = np.arange(L)
    indices = T[None, :] + (di[:, None] + filter_length * np.arange(N)[:, None])
    sinc_times = T - df[:, None] - (L - 1) / 2

    # we'll need to window also all the sincs at once
    windows = np.tile(np.hanning(L), N)

    # compute all sinc with one call
    bank_flat[indices.ravel()] = windows * np.sinc(sinc_times.ravel())

    return np.reshape(bank_flat, (N, -1))


def levinson(r, b):
    """

    Solve a system of the form Rx=b where R is hermitian toeplitz matrix and b
    is any vector using the generalized Levinson recursion as described in M.H.
    Hayes, Statistical Signal Processing and Modelling, p. 268.

    Parameters
    ----------
    r: 
        First column of R, toeplitz hermitian matrix.
    b: 
        The right-hand argument. If b is a matrix, the system is solved
        for every column vector in b.

    Returns
    -------
    numpy array
        The solution of the linear system Rx = b.
    """

    p = b.shape[0]

    a = np.array([1])
    x = b[np.newaxis, 0,] / r[0]
    epsilon = r[0]

    for j in np.arange(1, p):

        g = np.sum(np.conj(r[1 : j + 1]) * a[::-1])
        gamma = -g / epsilon
        a = np.concatenate((a, np.zeros(1))) + gamma * np.concatenate(
            (np.zeros(1), np.conj(a[::-1]))
        )
        epsilon = epsilon * (1 - np.abs(gamma) ** 2)
        delta = np.dot(np.conj(r[1 : j + 1]), np.flipud(x))
        q = (b[j,] - delta) / epsilon
        if len(b.shape) == 1:
            x = np.concatenate((x, np.zeros(1))) + q * np.conj(a[::-1])
        else:
            x = np.concatenate((x, np.zeros((1, b.shape[1]))), axis=0) + q * np.conj(
                a[::-1, np.newaxis]
            )

    return x


def autocorr(x, p, biased=True, method="numpy"):
    """
    Compute the autocorrelation for real signal `x` up to lag `p`.

    Parameters
    ----------
    x : numpy array
        Real signal in time domain.
    p : int
        Amount of lag. When solving for LPC coefficient, this is typically the
        LPC order.
    biased : bool
        Whether to return biased autocorrelation (default) or unbiased. As
        there are fewer samples for larger lags, the biased estimate tends to
        have better statistical stability under noise.
    method : 'numpy, 'fft', 'time', `pra`
        Method for computing the autocorrelation: in the frequency domain with
        `fft` or `pra` (`np.fft.rfft` is used so only real signals are
        supported), in the time domain with `time`, or with `numpy`'s built-in
        function `np.correlate` (default). For `p < log2(len(x))`, the time
        domain approach may be more efficient.

    Returns
    -------
    numpy array
        Autocorrelation for `x` up to lag `p`.
    """
    L = len(x)
    if method == "fft":
        X = np.fft.rfft(np.r_[x, np.zeros_like(x)])
        r = np.fft.irfft(X * np.conj(X))[: p + 1]
    elif method == "time":
        r = np.array([np.dot(x[: L - m], x[m:]) for m in range(p + 1)])
    elif method == "numpy":
        r = np.correlate(x, x, "full")[L - 1 : L + p]
    elif method == "pra":
        r = correlate(x, x)[L - 1 : L + p]
    else:
        raise ValueError(
            "Invalid `method` for computing autocorrelation"
            "choose one of: `fft`, `time`, `numpy`, `pra`."
        )

    if biased:
        return r / L
    else:
        return r / np.arange(L, L - p - 1, step=-1)


def lpc(x, p, biased=True):
    """
    Compute `p` LPC coefficients for a speech segment `x`.

    Parameters
    ----------
    x : numpy array
        Real signal in time domain.
    p : int
        Amount of lag. When solving for LPC coefficient, this is typically the
        LPC order.
    biased : bool
        Whether to use biased autocorrelation (default) or unbiased. As there
        are fewer samples for larger lags, the biased estimate tends to have
        better statistical stability under noise.

    Returns
    -------
    numpy array
        `p` LPC coefficients.
    """
    # compute autocorrelation
    r = autocorr(x, p, biased)

    # solve Yule-Walker equations for LPC coefficients
    return levinson(r[:p], r[1:])


def goertzel(x, k):
    """ Goertzel algorithm to compute DFT coefficients """

    N = x.shape[0]
    f = k / float(N)

    a = np.r_[1.0, -2.0 * np.cos(2.0 * np.pi * f), 1.0]
    b = np.r_[1]
    s = signal.lfilter(b, a, x)
    y = np.exp(2j * np.pi * f) * s[-1] - s[-2]

    return y


"""
GEOMETRY UTILITIES
"""


def angle_from_points(x1, x2):
    return np.angle((x1[0, 0] - x2[0, 0]) + 1j * (x1[1, 0] - x2[1, 0]))



'''
Least Mean Squares Family
=========================

Implementations of adaptive filters from the LMS class. These algorithms have a
low complexity and reliable behavior with a somewhat slower convergence.
'''
from __future__ import division, print_function, absolute_import
import numpy as np
import scipy.linalg as la

from .adaptive_filter import AdaptiveFilter

class NLMS(AdaptiveFilter):
    '''
    Implementation of the normalized least mean squares algorithm (NLMS)


    Parameters
    ----------
    length: int
        the length of the filter
    mu: float, optional
        the step size (default 0.5)
    '''
    
    def __init__(self, length, mu=0.5):
        
        self.mu = mu
        AdaptiveFilter.__init__(self, length)
        
    def update(self, x_n, d_n):
        '''
        Updates the adaptive filter with a new sample

        Parameters
        ----------
        x_n: float
            the new input sample
        d_n: float
            the new noisy reference signal
        '''
        
        AdaptiveFilter.update(self, x_n, d_n)
        
        e = self.d - np.inner(self.x, self.w)
        self.w += self.mu*e*self.x/np.inner(self.x,self.x)
        

class BlockLMS(NLMS):
    '''
    Implementation of the least mean squares algorithm (NLMS) in its block form

    Parameters
    ----------
    length: int
        the length of the filter
    mu: float, optional
        the step size (default 0.01)
    L: int, optional
        block size (default is 1)
    nlms: bool, optional
        whether or not to normalize as in NLMS (default is False)
    '''
    
    def __init__(self, length, mu=0.01, L=1, nlms=False):
        
        self.nlms = nlms
        
        # sketching parameters
        self.L = L # block size
        
        NLMS.__init__(self, length, mu=mu)
        self.reset()
        
    def reset(self):
        '''
        Reset the state of the adaptive filter
        '''
        NLMS.reset(self)
        # We need to redefine these two guys
        self.d = np.zeros((self.L))
        self.x = np.zeros((self.L+self.length-1))
        
    def update(self, x_n, d_n):
        '''
        Updates the adaptive filter with a new sample

        Parameters
        ----------
        x_n: float
            the new input sample
        d_n: float
            the new noisy reference signal
        '''
        
        # Update the internal buffers
        self.n += 1
        slot = self.L - ((self.n-1) % self.L) - 1
        self.x[slot] = x_n
        self.d[slot] = d_n
        
        # Block update
        if self.n % self.L == 0:
            
            # block-update parameters
            X = la.hankel(self.x[:self.L],r=self.x[self.L-1:])
            
            e = self.d - np.dot(X, self.w)
            
            if self.nlms:
                norm = np.linalg.norm(X, axis=1)**2
                if self.L == 1:
                    X = X/norm[0]
                else:
                    X = (X.T/norm).T
                
            self.w += self.mu*np.dot(X.T, e)
            
            # Remember a few values
            self.x[-self.length+1:] = self.x[0:self.length-1]
        



# -*- coding: utf-8 -*-
'''
Google's Speech Commands Dataset
================================
The Speech Commands Dataset has 65,000 one-second long utterances of 30 short 
words, by thousands of different people, contributed by members of the public
through the AIY website. It’s released under a Creative Commons BY 4.0 license.

More info about the dataset can be found at the link below:

https://research.googleblog.com/2017/08/launching-speech-commands-dataset.html

AIY website for contributing recordings:

https://aiyprojects.withgoogle.com/open_speech_recording

Tutorial on creating a word classifier:

https://www.tensorflow.org/versions/master/tutorials/audio_recognition
'''

import os, glob
import numpy as np
from scipy.io import wavfile

try:
    import sounddevice as sd
    have_sounddevice = True
except:
    have_sounddevice = False

from .utils import download_uncompress
from .base import Meta, AudioSample, Dataset

url = "http://download.tensorflow.org/data/speech_commands_v0.01.tar.gz"


class GoogleSpeechCommands(Dataset):
    '''
    This class will load the Google Speech Commands Dataset in a
    structure that is convenient to be processed.

    Attributes
    ----------
    basedir: str
        The directory where the Speech Commands Dataset is located/downloaded.
    size_by_samples: dict
        A dictionary whose keys are the words in the dataset. The values are the number of occurances for that particular word.
    subdirs: list
        The list of subdirectories in ``basedir``, where each sound type is the name of a subdirectory.
    classes: list
        The list of all sounds, same as the keys of ``size_by_samples``.

    Parameters
    ----------
    basedir: str, optional
        The directory where the Google Speech Commands dataset is located/downloaded. By default, this is the current directory.
    download: bool, optional
        If the corpus does not exist, download it.
    build: bool, optional
        Whether or not to build the dataset. By default, it is.
    subset: int, optional
        Build a dataset that contains all noise samples and `subset` samples per word. By default, the dataset will be built with all samples.
    seed: int, optional
        Which seed to use for the random generator when selecting a subset of samples. By default, ``seed=0``.
    '''

    def __init__(self, basedir=None, download=False, build=True, subset=None,
        seed=0, **kwargs):

        # initialize
        Dataset.__init__(self)
        self.size_by_samples = {}

        # default base directory is the current one
        self.basedir = basedir
        if basedir is None:
            self.basedir = './google_speech_commands'

        # check the directory exists and download otherwise
        if not os.path.exists(self.basedir):
            if download:
                print('Downloading', url, 'into', self.basedir, '...')
                download_uncompress(url=url, path=self.basedir)
            else:
                raise ValueError('Dataset directory does not exist. Create or set download option.')
        else:
            print("Dataset exists! Using %s" % self.basedir)

        # set seed of random number generator
        self.seed = seed
        self.rng = np.random.RandomState(seed)

        if build:
            self.build_corpus(subset, **kwargs)


    def build_corpus(self, subset=None, **kwargs):
        '''
        Build the corpus with some filters (speech or not speech, sound type).
        '''

        self.subdirs = glob.glob(os.path.join(self.basedir,'*','.'))
        self.classes = [s.split(os.sep)[-2] for s in self.subdirs]

        # go through all subdirectories / soundtypes
        for idx, word in enumerate(self.classes):

            if word == '_background_noise_':
                speech = False
            else:
                speech = True

            # get all list of all files in the subdirectory
            word_path = self.subdirs[idx]
            files = glob.glob(os.path.join(word_path, '*.wav'))

            # for speech files, select desired subset
            if subset and speech:
                rand_idx = np.arange(len(files)) 
                n_files = min(subset, len(files))
                self.rng.shuffle(rand_idx)
                files = [files[i] for i in rand_idx[:n_files]]
            
            self.size_by_samples[word] = len(files)

            # add each file to the corpus
            for filename in files:

                file_loc = os.path.join(self.basedir, word, os.path.basename(filename))

                # could also add score of original model for each word?
                if speech:
                    meta = Meta(word=word, speech=speech, file_loc=file_loc)
                else:
                    noise_type = os.path.basename(filename).split(".")[0]
                    meta = Meta(word="NA", noise_type=noise_type, speech=speech, file_loc=file_loc)

                if meta.match(**kwargs):
                    self.add_sample(GoogleSample(filename, **meta.as_dict()))


    def filter(self, **kwargs):
        '''
        Filter the dataset and select samples that match the criterias provided
        The arguments to the keyword can be 1) a string, 2) a list of strings, 3)
        a function. There is a match if one of the following is True.

        1. ``value == attribute``
        2. ``value`` is a list and ``attribute in value == True``
        3. ``value`` is a callable (a function) and ``value(attribute) == True``
        '''

        # first, create the new empty corpus
        new_corpus = GoogleSpeechCommands(basedir=self.basedir, build=False,
            seed=self.seed)

        # finally, add all the sentences
        for s in self.samples:
            new_corpus.add_sample_matching(s, **kwargs)

        return new_corpus


class GoogleSample(AudioSample):
    '''
    Create the sound object.

    Parameters
    ----------
    path: str
      the path to the audio file
    **kwargs:
      metadata as a list of keyword arguments

    Attributes
    ----------
    data: array_like
      the actual audio signal
    fs: int
      sampling frequency
    '''

    def __init__(self,path,**kwargs):
        '''
        Create the the sound object
        path: string
          the path to a particular sample
        '''

        fs,data = wavfile.read(path)
        AudioSample.__init__(self, data, fs, **kwargs)

    def __str__(self):
        '''string representation'''

        if self.meta.speech:
            template = 'speech: ''{speech}''; word: ''{word}''; file_loc: ''{file_loc}'''
        else:
            template = 'speech: ''{speech}''; noise type: ''{noise_type}''; file_loc: ''{file_loc}'''
        s = template.format(**self.meta.as_dict())
        return s


    def plot(self,**kwargs):
        '''Plot the spectogram'''
        try:
            import matplotlib.pyplot as plt 
        except ImportError:
            print('Warning: matplotlib is required for plotting')
            return
        AudioSample.plot(self,**kwargs)
        if self.meta.speech:
            plt.title(self.meta.file_loc)
        else:
            plt.title(self.meta.file_loc)



# Test of the CCW3P routine
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.

import numpy as np
import pyroomacoustics as pra

cases = {
        'anti-clockwise' : {
            'points' : np.array([
                [1,-1], [2,-1], [1,0]
                ]),
            'expected' : 1,  # anti-clockwise
            'label' : 'Test: CCW3P anti-clockwise',
            },
        'clockwise' : {
            'points' : np.array([
                [1,-1], [1,0], [2,-1]
                ]),
            'expected' : -1,  # clockwise
            'label' : 'Test: CCW3P clockwise',
            },
        'co-linear' : {
            'points' : np.array([
                [0,0], [0.5,0.5], [1,1]
                ]),
            'expected' : 0,  # co-linear
            'label' : 'Test: CCW3P co-linear',
            },
        }

def ccw3p(case):

    p1, p2, p3 = case['points']

    r = pra.libroom.ccw3p(p1, p2, p3)

    assert r == case['expected'], (case['label']
            + ' returned: {}, expected {}'.format(r, case['expected']))

def test_ccw3p_anticlockwise():
    ccw3p(cases['anti-clockwise'])

def test_ccw3p_clockwise():
    ccw3p(cases['clockwise'])

def test_ccw3p_colinear():
    ccw3p(cases['co-linear'])

if __name__ == '__main__':

    for lbl, case in cases.items():

        try:
            ccw3p(case)
        except:
            print('{} failed'.format(lbl))



'''
The CMU ARCTIC Dataset
======================

The CMU_ARCTIC databases were constructed at the Language Technologies
Institute at Carnegie Mellon University as phonetically balanced, US English
single speaker databases designed for unit selection speech synthesis research.
A detailed report on the structure and content of the database and the
recording environment etc is available as a Carnegie Mellon University,
Language Technologies Institute Tech Report CMU-LTI-03-177 and is also
available here.

The databases consist of around 1150 utterances carefully selected from
out-of-copyright texts from Project Gutenberg. The databses include US English
male (bdl) and female (slt) speakers (both experinced voice talent) as well as
other accented speakers.

The 1132 sentence prompt list is available from cmuarctic.data

The distributions include 16KHz waveform and simultaneous EGG signals. Full
phoentically labelling was perfromed by the CMU Sphinx using the FestVox based
labelling scripts. Complete runnable Festival Voices are included with the
database distributions, as examples though better voices can be made by
improving labelling etc.

License: Permissive, attribution required

Price: Free

URL: http://www.festvox.org/cmu_arctic/
'''
import os
import numpy as np
from scipy.io import wavfile

try:
    import sounddevice as sd
    have_sounddevice = True
except:
    have_sounddevice = False

from .utils import download_uncompress
from .base import Meta, AudioSample, Dataset

# The speakers codes and attributes
cmu_arctic_speakers = {
        'aew' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'US' },
        'ahw' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'German' },
        'aup' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'Indian' },
        'awb' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'Scottish' },
        'axb' : { 'sex' : 'female', 'lang' : 'US English', 'accent' : 'Indian' },
        'bdl' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'US' },
        'clb' : { 'sex' : 'female', 'lang' : 'US English', 'accent' : 'US' },
        'eey' : { 'sex' : 'female', 'lang' : 'US English', 'accent' : 'US' },
        'fem' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'Irish' },
        'gka' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'Indian' },
        'jmk' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'Canadian' },
        'ksp' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'Indian' },
        'ljm' : { 'sex' : 'female', 'lang' : 'US English', 'accent' : 'US' },
        'lnh' : { 'sex' : 'female', 'lang' : 'US English', 'accent' : 'US' },
        'rms' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'US' },
        'rxr' : { 'sex' : 'male',   'lang' : 'US English', 'accent' : 'Dutch' },
        'slp' : { 'sex' : 'female', 'lang' : 'US English', 'accent' : 'Indian' },
        'slt' : { 'sex' : 'female', 'lang' : 'US English', 'accent' : 'US' },
        }

# The sentences in the database
cmu_arctic_sentences = {}

# Directory structure
speaker_dir = 'cmu_us_{}_arctic'

# Download info
url_base = 'http://festvox.org/cmu_arctic/packed/{}.tar.bz2'.format(speaker_dir)

class CMUArcticCorpus(Dataset):
    '''
    This class will load the CMU ARCTIC corpus in a
    structure amenable to be processed.

    Attributes
    ----------
    basedir: str, option
        The directory where the CMU ARCTIC corpus is located/downloaded. By
        default, this is the current directory.
    info: dict
        A dictionary whose keys are the labels of metadata fields attached to the samples.
        The values are lists of all distinct values the field takes.
    sentences: list of CMUArcticSentence
        The list of all utterances in the corpus

    Parameters
    ----------
    basedir: str, optional
        The directory where the CMU ARCTIC corpus is located/downloaded. By
        default, this is the current directory.
    download: bool, optional
        If the corpus does not exist, download it.
    speaker: str or list of str, optional
        A list of the CMU ARCTIC speakers labels. If provided, only
        those speakers are loaded. By default, all speakers are loaded.
    sex: str or list of str, optional
        Can be 'female' or 'male'
    lang: str or list of str, optional
        The language, only 'English' is available here
    accent: str of list of str, optional
        The accent of the speaker
    '''


    def __init__(self, basedir=None, download=False, build=True, **kwargs):

        # initialize
        Dataset.__init__(self)

        # we give a meaningful alias to the sample list from the base class
        self.sentences = self.samples

        # default base directory is the current one
        self.basedir = basedir
        if basedir is None:
            self.basedir = './CMU_ARCTIC'

        # if no speaker is specified, use all the speakers
        if 'speaker' not in kwargs:
            kwargs['speaker'] = list(cmu_arctic_speakers.keys())

        # we need this to know which speakers to download
        speakers = kwargs['speaker']
        if not isinstance(speakers, list):
            speakers = [speakers]

        # check everything requested is there or download
        if not os.path.exists(self.basedir):
            if download:
                os.mkdir(self.basedir)
            else:
                raise ValueError('Corpus directory does not exist. Create or set download option.')

        # remove invalid speaker keys
        n_speakers = len(speakers)
        speakers = [speaker for speaker in speakers if speaker in cmu_arctic_speakers.keys()]
        if n_speakers != len(speakers):
            import warnings
            warnings.warn('Some invalid speakers were removed from the list.', RuntimeWarning)


        # now crawl the speakers directories, download when necessary
        for speaker in speakers:
            sdir = os.path.join(self.basedir, speaker_dir.format(speaker))

            # check the directory exists and download otherwise
            if not os.path.exists(sdir):
                if download:
                    url = url_base.format(speaker)
                    print('Download', url, 'into', self.basedir, '...')
                    download_uncompress(url, path=self.basedir)
                else:
                    raise ValueError('Missing speaker directory. Please download.')

        # now, we populate the sentence data structure containing the list
        # of all distinct sentences spoken in the database
        for speaker in speakers:
            # Now crawl the directory to get all the data
            sdir = os.path.join(self.basedir, speaker_dir.format(speaker))

            all_files = []

            with open(os.path.join(sdir, 'etc/txt.done.data'), 'r') as f:
                for line in f.readlines():
                    # extract the file path
                    tag = line.split(' ')[1]
                    path = os.path.join(sdir, 'wav/' + tag + '.wav')
                    text = line.split('"')[1]

                    # now shorten the tag for internal use
                    tag = tag[-5:]

                    # add the dict of sentences
                    if tag not in cmu_arctic_sentences:
                        cmu_arctic_sentences[tag] = { 
                                'text': text, 
                                'paths' : { speaker : path } 
                                }
                    else:
                        cmu_arctic_sentences[tag]['paths'][speaker] = path

        if build:
            self.build_corpus(**kwargs)

    def build_corpus(self, **kwargs):
        '''
        Build the corpus with some filters (sex, lang, accent, sentence_tag, sentence)
        '''

        # Check all the sentences
        for tag, info in cmu_arctic_sentences.items():

            # And all speakers for each sentence
            for speaker, path in info['paths'].items():

                # This is the metadata for this sample
                meta = Meta(speaker=speaker, tag=tag, text=info['text'], **cmu_arctic_speakers[speaker])

                # it there is a match, add it
                # The reason we do the match before creating the Sentence object is that
                # we don't want to read the file from disk if there is no match
                if meta.match(**kwargs):
                    self.add_sample(CMUArcticSentence(path, **meta.as_dict()))

    def filter(self, **kwargs):
        '''
        Filter the corpus and selects samples that match the criterias provided
        The arguments to the keyword can be 1) a string, 2) a list of strings, 3)
        a function. There is a match if one of the following is True.

        1. ``value == attribute``
        2. ``value`` is a list and ``attribute in value == True``
        3. ``value`` is a callable (a function) and ``value(attribute) == True``
        '''

        # first, create the new empty corpus
        new_corpus = CMUArcticCorpus(basedir=self.basedir, build=False, speaker=[])

        # finally, add all the sentences
        for s in self.samples:
            new_corpus.add_sample_matching(s, **kwargs)

        return new_corpus


class CMUArcticSentence(AudioSample):
    '''
    Create the sentence object

    Parameters
    ----------
    path: str
        the path to the audio file
    **kwargs:
        metadata as a list of keyword arguments

    Attributes
    ----------
    data: array_like
        The actual audio signal
    fs: int
        sampling frequency
    '''

    def __init__(self, path, **kwargs):
        '''
        Create the sentence object

        path: (string)
            the path to the particular sample
        '''

        # Read in the wav file
        fs, data = wavfile.read(path)
        
        # initialize the parent object
        AudioSample.__init__(self, data, fs, **kwargs)


    def __str__(self):
        ''' String representation '''
        template = '{speaker} ({sex}, {lang}/{accent}); {tag}: ''{text}'''
        s = template.format(**self.meta.as_dict())
        return s

    def plot(self, **kwargs):
        ''' Plot the spectrogram '''
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            print('Warning: matplotlib is required for plotting')
            return
        AudioSample.plot(self, **kwargs)
        plt.title(self.meta.text)




from __future__ import division

import numpy as np

try:
    import mkl_fft._numpy_fft as fft
except:
    from numpy import fft

def tdoa(signal, reference, interp=1, phat=False, fs=1, t_max=None):
    '''
    Estimates the shift of array signal with respect to reference
    using generalized cross-correlation

    Parameters
    ----------
    signal: array_like
        The array whose tdoa is measured
    reference: array_like
        The reference array
    interp: int, optional
        The interpolation factor for the output array, default 1.
    phat: bool, optional
        Apply the PHAT weighting (default False)
    fs: int or float, optional
        The sampling frequency of the input arrays, default=1

    Returns
    -------
    The estimated delay between the two arrays
    '''

    signal = np.array(signal)
    reference = np.array(reference)

    N1 = signal.shape[0]
    N2 = reference.shape[0]

    r_12 = correlate(signal, reference, interp=interp, phat=phat)

    delay = (np.argmax(np.abs(r_12)) / interp  - (N2 - 1) ) / fs

    return delay

def correlate(x1, x2, interp=1, phat=False):
    '''
    Compute the cross-correlation between x1 and x2

    Parameters
    ----------
    x1,x2: array_like
        The data arrays
    interp: int, optional
        The interpolation factor for the output array, default 1.
    phat: bool, optional
        Apply the PHAT weighting (default False)

    Returns
    -------
    The cross-correlation between the two arrays
    '''

    N1 = x1.shape[0]
    N2 = x2.shape[0]

    N = N1 + N2 - 1

    X1 = fft.rfft(x1, n=N)
    X2 = fft.rfft(x2, n=N)

    if phat:
        eps1 = np.mean(np.abs(X1)) * 1e-10
        X1 /= (np.abs(X1) + eps1)
        eps2 = np.mean(np.abs(X2)) * 1e-10
        X2 /= (np.abs(X2) + eps2)

    m = np.minimum(N1, N2)

    out = fft.irfft(X1*np.conj(X2), n=int(N*interp))

    return np.concatenate([out[-interp*(N2-1):], out[:(interp*N1)]])

def delay_estimation(x1, x2, L):
    '''
    Estimate the delay between x1 and x2.
    L is the block length used for phat
    '''

    K = int(np.minimum(x1.shape[0], x2.shape[0]) / L)

    delays = np.zeros(K)
    for k in range(K):
        delays[k] = tdoa(x1[k*L:(k+1)*L], x2[k*L:(k+1)*L], phat=True)

    return int(np.median(delays))


def time_align(ref, deg, L=4096):
    '''
    return a copy of deg time-aligned and of same-length as ref.
    L is the block length used for correlations.
    '''

    # estimate delay of signal
    from numpy import zeros, minimum
    delay = delay_estimation(ref, deg, L)

    # time-align with reference segment for error metric computation
    sig = zeros(ref.shape[0])
    if (delay >= 0):
        length = minimum(deg.shape[0], ref.shape[0]-delay)
        sig[delay:length+delay] = deg[:length]
    else:
        length = minimum(deg.shape[0]+delay, ref.shape[0])
        sig = zeros(ref.shape)
        sig[:length] = deg[-delay:-delay+length]

    return sig




'''
The TIMIT Dataset
=================

The TIMIT corpus of read speech is designed to provide speech data for acoustic-phonetic studies and for the development and evaluation of automatic speech recognition systems. TIMIT contains broadband recordings of 630 speakers of eight major dialects of American English, each reading ten phonetically rich sentences. The TIMIT corpus includes time-aligned orthographic, phonetic and word transcriptions as well as a 16-bit, 16kHz speech waveform file for each utterance. Corpus design was a joint effort among the Massachusetts Institute of Technology (MIT), SRI International (SRI) and Texas Instruments, Inc. (TI). The speech was recorded at TI, transcribed at MIT and verified and prepared for CD-ROM production by the National Institute of Standards and Technology (NIST).

The TIMIT corpus transcriptions have been hand verified. Test and training subsets, balanced for phonetic and dialectal coverage, are specified. Tabular computer-searchable information is included as well as written documentation.

Unfortunately, this is a proprietary dataset. A licensed can be obtained for $125 to $250 depending on your
status (academic or otherwise).

**Deprecation Warning:** The interface of TimitCorpus will change in the near
future to match that of :py:obj:`pyroomacoustics.datasets.cmu_arctic.CMUArcticCorpus`

URL: https://catalog.ldc.upenn.edu/ldc93s1
'''

import os
import numpy as np

from pyroomacoustics import stft

try:
    import sounddevice as sd
    have_sounddevice = True
except:
    have_sounddevice = False


def _read_nist_wav(filename):
    '''
    A custom implementation of wav reader that let's read the NIST files from TIMIT

    Decode wav header and create numpy array out of raw data.

    Parameters
    ----------
    filename: str
        the name of the file to decode
    '''
    _byte_format = { 2 : np.int16, 4 : np.int32 }  # helper

    # get raw data
    f = open(filename, 'rb')
    file_data = f.read()
    f.close()

    file_type = file_data[:8].decode()
    if file_type != 'NIST_1A\n':
        raise ValueError('Format might be wrong. We need a NIST_1A file.')

    # separate  header/data
    header_size = int(file_data[8:16].decode())
    header = file_data[16:header_size].decode()
    data = file_data[header_size:]


    # extract all the key/values for all fields of the header
    fields = [line.split() for line in header.split('\n')]

    header_dict = dict()
    for field in fields:
        if len(field) != 3:
            continue

        key = field[0]
        t = field[1][1]

        if t == 'i':
            val = int(field[2])
        elif t == 's':
            val = field[2]
        elif t == 'r':
            val = float(field[2])
        else:
            raise ValueError('Unknown type field for NIST header')

        header_dict[key] = val

    # extract parameters from header
    nch = header_dict['channel_count']
    fs = header_dict['sample_rate']
    bytes_per_sample = header_dict['sample_n_bytes']
    try:
        dt = _byte_format[bytes_per_sample]
    except KeyError:
        raise ValueError("Currently we only support 16 or 32 bits files. Sorry")

    # Build the array
    array = np.frombuffer(data, dtype=_byte_format[bytes_per_sample])
    if nch > 1:
        array = array.reshape((-1,nch))

    return fs, array


class Word:
    '''
    A class used for words of the TIMIT corpus

    Attributes
    ----------
    word: str
        The spelling of the word
    boundaries: list
        The limits of the word within the sentence
    samples: array_like
        A view on the sentence samples containing the word
    fs: int
        The sampling frequency
    phonems: list
        A list of phones contained in the word
    features: array_like
        A feature array (e.g. MFCC coefficients)

    Parameters
    ----------
    word: str
        The spelling of the word
    boundaries: list
        The limits of the word within the sentence
    data: array_like
        The nd-array that contains all the samples of the sentence
    fs: int
        The sampling frequency
    phonems: list, optional
        A list of phones contained in the word

    '''


    def __init__(self, word, boundaries, data, fs, phonems=None):

        self.word = word
        self.phonems = phonems
        self.boundaries = boundaries
        self.samples = data[boundaries[0]:boundaries[1]]
        self.fs = fs
        self.features = None

    def __str__(self):
        return self.word

    def plot(self):
        try:
            import matplotlib.pyplot as plt
            import seaborn as sns
        except ImportError:
            return

        sns.set_style('white')

        L = self.samples.shape[0]
        plt.plot(np.arange(L)/self.fs, self.samples)
        plt.xlim((0,L/self.fs))
        plt.xlabel('Time')
        plt.title(self.word)

    def play(self):
        ''' Play the sound sample '''
        if have_sounddevice:
            sd.play(self.samples, samplerate=self.fs)
        else:
            print('Warning: sounddevice package is required to play audiofiles.')

    def mfcc(self, frame_length=1024, hop=512):
        ''' compute the mel-frequency cepstrum coefficients of the word samples '''
        self.features = mfcc(self.samples, L=frame_length, hop=hop)


class Sentence:
    '''
    Create the sentence object

    Parameters
    ----------
    path: (string)
        the path to the particular sample

    Attributes
    ----------
    speaker: str
        Speaker initials
    id: str
        a digit to disambiguate identical initials
    sex: str
        Speaker gender (M or F)
    dialect: str
        Speaker dialect region number:

        1. New England
        2. Northern
        3. North Midland
        4. South Midland
        5. Southern
        6. New York City
        7. Western
        8. Army Brat (moved around)
    fs: int
        sampling frequency
    samples: array_like (n_samples,)
        the audio track
    text: str
        the text of the sentence
    words: list
        list of Word objects forming the sentence
    phonems: list
        List of phonems contained in the sentence. Each element is a
        dictionnary containing a 'bnd' with the limits of the phonem and 'name'
        that is the phonem transcription.
    '''

    def __init__(self, path):
        '''
        Create the sentence object

        path: (string)
            the path to the particular sample
        '''

        path, ext = os.path.splitext(path)

        t = path.split('/')

        # extract the attributes
        self.dialect = t[-3]
        self.sex = t[-2][0]
        self.speaker = t[-2][1:5]
        self.id = t[-1]

        # Read in the wav file
        self.fs, self.data = _read_nist_wav(path + '.WAV')
        self.samples = self.data

        # Read the sentence text
        f = open(path + '.TXT', 'r')
        lines = f.readlines()
        self.text = ' '.join(lines[0].split()[2:])
        f.close()

        # Read the word list
        self.words = []
        self.phonems = []

        # open the word file
        f = open(path + '.WRD', 'r')
        w_lines = f.readlines()
        f.close()

        # get all lines from the phonem file
        f_ph = open(path + '.PHN', 'r')
        ph_lines = f_ph.readlines()
        ph_l_index = 0
        f_ph.close()

        for line in w_lines:
            t = line.split()

            # just a sanity check
            if len(t) == 3:

                # the word boundary
                w_bnd = np.array([int(t[0]), int(t[1])])

                # recover the phonems making up the word
                w_ph_list = []
                while ph_l_index < len(ph_lines):

                    ph_line = ph_lines[ph_l_index]
                    u = ph_line.split()

                    # phonem boundary
                    ph_bnd = np.array([int(u[0]), int(u[1])])

                    # Check phonem boundary does not exceeds word boundary
                    if ph_bnd[1] > w_bnd[1]:
                        break

                    # add to sentence object phonems list
                    self.phonems.append({'name':u[2], 'bnd':ph_bnd})

                    # increase index
                    ph_l_index += 1

                    # now skip until beginning of word
                    if ph_bnd[0] < w_bnd[0]:
                        continue

                    # add phonem to word if (with adjusted boundaries wrt to start of word)
                    w_ph_list.append({'name':u[2], 'bnd':ph_bnd - w_bnd[0]})

                # Finally create word object
                self.words.append(Word(t[2], w_bnd, self.data, self.fs, phonems=w_ph_list))

        # Read the remaining phonem(s)
        while ph_l_index < len(ph_lines):
            ph_line = ph_lines[ph_l_index]
            u = ph_line.split()

            if len(u) == 3:
                # phonem boundary
                ph_bnd = (int(u[0]), int(u[1]))

                # add to sentence object phonems list
                self.phonems.append({'name':u[2], 'bnd':ph_bnd})

            ph_l_index += 1


    def __str__(self):
        s = " ".join([self.dialect, self.sex, self.speaker, self.id, self.text])
        return s

    def play(self):
        ''' Play the sound sample '''
        if have_sounddevice:
            sd.play(self.data, samplerate=self.fs)
        else:
            print('Warning: sounddevice package is required to play audiofiles.')

    def plot(self, L=512, hop=128, zpb=0, phonems=False, **kwargs):

        try:
            import matplotlib.pyplot as plt
            import seaborn as sns
        except ImportError:
            return

        sns.set_style('white')
        X = stft(self.data, L=L, hop=hop, zp_back=zpb, transform=np.fft.rfft, win=np.hanning(L+zpb))
        X = 10*np.log10(np.abs(X)**2).T

        plt.imshow(X, origin='lower', aspect='auto')

        ticks = []
        ticklabels = []

        if phonems:
            for phonem in self.phonems:
                plt.axvline(x=phonem['bnd'][0]/hop)
                plt.axvline(x=phonem['bnd'][1]/hop)
                ticks.append((phonem['bnd'][1]+phonem['bnd'][0])/2/hop)
                ticklabels.append(phonem['name'])

        else:
            for word in self.words:
                plt.axvline(x=word.boundaries[0]/hop)
                plt.axvline(x=word.boundaries[1]/hop)
                ticks.append((word.boundaries[1]+word.boundaries[0])/2/hop)
                ticklabels.append(word.word)

        plt.xticks(ticks, ticklabels, rotation=-45)
        plt.yticks([],[])
        plt.tick_params(axis='both', which='major', labelsize=14)


class TimitCorpus:
    '''
    TimitCorpus class

    Parameters
    ----------
    basedir : (string)
        The location of the TIMIT database
    directories : (list of strings)
        The subdirectories containing the data (['TEST','TRAIN'])
    sentence_corpus : (dict)
        A dictionnary that contains a list of Sentence objects for each sub-directory
    word_corpus : (dict)
        A dictionnary that contains a list of Words objects for each sub-directory
        and word available in the corpus
    '''

    def __init__(self, basedir):
        ''' Initialize basic attributes of the class '''

        import warnings
        warnings.warn("This interface for TIMIT is deprecated "
                    + "and will be replaced soon to match the "
                    + "Dataset base class and the CMUArcticCorpus "
                    + "interface.", FutureWarning)

        if not os.path.exists(basedir):
            raise ValueError('The directory ''{}'' does not exist.'.format(basedir))

        if not os.path.exists(basedir + '/TEST') or not os.path.exists(basedir + '/TRAIN'):
            raise ValueError('The directory ''{}'' does not contain sub-directories TEST and TRAIN'.format(basedir))

        self.basedir = basedir
        self.directories = ['TEST','TRAIN']
        self.sentence_corpus = None
        self.word_corpus = None


    def build_corpus(self, sentences=None, dialect_region=None, speakers=None, sex=None):
        '''
        Build the corpus

        The TIMIT database structure is encoded in the directory sturcture:

        basedir
            TEST/TRAIN
                Regional accent index (1 to 8)
                    Speakers (one directory per speaker)
                        Sentences (one file per sentence)

        Parameters
        ----------
        sentences: (list)
            A list containing the sentences to which we want to restrict the corpus
            Example: sentences=['SA1','SA2']
        dialect_region: (list of int)
            A list to which we restrict the dialect regions
            Example: dialect_region=[1, 4, 5]
        speakers: (list)
            A list of speakers acronym to which we want to restrict the corpus
            Example: speakers=['AKS0']
        sex: (string)
            Restrict to a single sex: 'F' for female, 'M' for male
        '''
        self.sentence_corpus = dict(zip(self.directories, [[],[]]))
        self.word_corpus = dict(zip(self.directories, [{},{}]))

        if dialect_region is not None:
            dialect_region = ['DR' + str(i) for i in dialect_region]

        # Read in all the sentences making use of TIMIT special directory structure
        for d in self.directories:
            dir1 = os.path.join(self.basedir, d)
            for dialect in next(os.walk(dir1))[1]:
                # check if dialect is in exclusion list
                if dialect_region is not None and dialect not in dialect_region:
                    continue

                dir2 = os.path.join(dir1, dialect)

                for speaker in next(os.walk(dir2))[1]:
                    # check if sex is matching
                    if sex is not None and speaker[0] != sex:
                        continue
                    # check if speaker is not in exclusion list
                    if speakers is not None and speaker not in speakers:
                        continue

                    dir3 = os.path.join(dir2, speaker)

                    for fil in os.listdir(dir3):
                        # just look at wav files to avoid duplicates
                        if fil.endswith('.WAV'):
                            sentence = os.path.splitext(fil)[0]
                            # check if sentence should be excluded
                            if sentences is not None and sentence not in sentences:
                                continue

                            # Create a new sentence object
                            path = os.path.join(dir3, sentence)
                            self.sentence_corpus[d].append(Sentence(path))

                            # Now add the words to the word corpus
                            for w in self.sentence_corpus[d][-1].words:
                                if not w.word in self.word_corpus[d]:
                                    self.word_corpus[d][w.word] = [w]
                                else:
                                    self.word_corpus[d][w.word].append(w)


    def get_word(self, d, w, index=0):
        ''' return instance index of word w from group (test or train) d '''
        return self.word_corpus[d][w][index]




"""
Experimental
============

A bunch of routines useful when doing measurements and experiments.
"""

__all__ = [
    "measure_ir",
    "physics",
    "point_cloud",
    "delay_calibration",
    "deconvolution",
    "localization",
    "signals",
    "rt60",
]

from .measure_ir import measure_ir
from .physics import calculate_speed_of_sound
from .point_cloud import PointCloud
from .delay_calibration import DelayCalibration
from .deconvolution import deconvolve, wiener_deconvolve
from .localization import tdoa, tdoa_loc, edm_line_search
from .signals import window, exponential_sweep, linear_sweep
from .rt60 import measure_rt60



# Single Channel Noise Removal using Iterative Wiener Filtering
# Copyright (C) 2019  Eric Bezzam, Laurent Colbois, Lionel Desarzens
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.

import numpy as np
from scipy import integrate
from pyroomacoustics import lpc


class IterativeWiener(object):
    """
    A class for performing **single channel** noise reduction in the frequency
    domain with a Wiener filter that is iteratively computed. This
    implementation is based off of the approach presented in:

        J. Lim and A. Oppenheim, *All-Pole Modeling of Degraded Speech,*
        IEEE Transactions on Acoustics, Speech, and Signal Processing 26.3
        (1978): 197-210.

    For each frame, a Wiener filter of the following form is computed and
    applied to the noisy samples in the frequency domain:

    .. math::

        H(\\omega) = \dfrac{P_S(\\omega)}{P_S(\\omega) + \\sigma_d^2},


    where :math:`P_S(\omega)` is the speech power spectral density and
    :math:`\sigma_d^2` is the noise variance.

    The following assumptions are made in order to arrive at the above filter
    as the optimal solution:

    - The noisy samples :math:`y[n]` can be written as:

      .. math::

        y[n] = s[n] + d[n],

      where :math:`s[n]` is the desired signal and :math:`d[n]` is the
      background noise.
    - The signal and noise are uncorrelated.
    - The noise is white Gaussian, i.e. it has a flat power spectrum with
      amplitude :math:`\sigma_d^2`.

    Under these assumptions, the above Wiener filter minimizes the mean-square
    error between the true samples :math:`s[0:N-1]` and the estimated one
    :math:`\hat{s[0:N-1]}` by filtering :math:`y[0:N-1]` with the above filter
    (with :math:`N` being the frame length).

    The fundamental part of this approach is correctly (or as well as possible)
    estimating the speech power spectral density :math:`P_S(\omega)` and the
    noise variance :math:`\sigma_d^2`. For this, we need a **voice activity
    detector** in order to determine when we have incoming speech. In this
    implementation, we use a simple energy threshold on the input frame, which
    is set with the `thresh` input parameter.

    **When no speech is identified**, the input frame is used to update the
    noise variance :math:`\sigma_d^2`. We could simply set :math:`\sigma_d^2`
    to the energy of the input frame. However, we employ a simple IIR filter in
    order to avoid abrupt changes in the noise level (thus adding an assumption
    of stationary):

    .. math::

        \sigma_d^2[k] = \\alpha \cdot \sigma_d^2[k-1] + (1-\\alpha) \cdot \sigma_y^2,

    where :math:`\\alpha` is the smoothing parameter and :math:`\sigma_y^2` is
    the energy of the input frame. A high value of :math:`\\alpha` will update
    the noise level very slowly, while a low value will make it very sensitive
    to changes at the input. The value for :math:`\\alpha` can be set with the
    `alpha` parameter.

    **When speech is identified in the input frame**, an iterative procedure is
    employed in order to estimate :math:`P_S(\omega)` (and therefore the Wiener
    filter :math:`H` as well). This procedure consists of computing :math:`p`
    `linear predictive coding (LPC) coefficients <https://en.wikipedia.org/wiki/Linear_predictive_coding>`_
    of the input frame. The number of LPC coefficients is set with the
    parameter `lpc_order`. These LPC coefficients form an all-pole filter that
    models the vocal tract as described in the above paper (Eq. 1). With these
    coefficients, we can then obtain an estimate of the speech power spectral
    density (Eq. 41b) and thus the corresponding Wiener filter (Eq. 41a). This
    Wiener filter is used to denoise the input frame. Moreover, with this
    denoised frame, we can compute new LPC coefficients and therefore a new
    Wiener filter. The idea behind this approach is that by iteratively
    computing the LPC coefficients as such, we can obtain a better estimate of
    the speech power spectral density. The number of iterations can be set with
    the `iterations` parameter.

    Below is an example of how to use this class to emulate a streaming/online
    input. A full example can be found
    `here <https://github.com/LCAV/pyroomacoustics/blob/master/examples/noise_reduction_wiener_filtering.py>`_.


    ::

        # initialize STFT and IterativeWiener objects
        nfft = 512
        stft = pra.transform.STFT(nfft, hop=nfft//2,
                                  analysis_window=pra.hann(nfft))
        scnr = IterativeWiener(frame_len=nfft, lpc_order=20, iterations=2,
                               alpha=0.8, thresh=0.01)

        # apply block-by-block
        for n in range(num_blocks):

            # go to frequency domain, 50% overlap
            stft.analysis(mono_noisy)

            # compute wiener output
            X = scnr.compute_filtered_output(
                    current_frame=stft.fft_in_buffer,
                    frame_dft=stft.X)

            # back to time domain
            mono_denoised = stft.synthesis(X)


    There also exists a "one-shot" function.

    ::

        # import or create `noisy_signal`
        denoised_signal = apply_iterative_wiener(noisy_signal, frame_len=512,
                                                 lpc_order=20, iterations=2,
                                                 alpha=0.8, thresh=0.01)


    Parameters
    ----------
    frame_len : int
        Frame length in samples.
    lpc_order : int
        Number of LPC coefficients to compute
    iterations : int
        How many iterations to perform in updating the Wiener filter for each
        signal frame.
    alpha : int
        Smoothing factor within [0,1] for updating noise level. Closer to `1`
        gives more weight to the previous noise level, while closer to `0`
        gives more weight to the current frame's level. Closer to `0` can track
        more rapid changes in the noise level. However, if a speech frame is
        incorrectly identified as noise, you can end up removing desired
        speech.
    thresh : float
        Threshold to distinguish between (signal+noise) and (noise) frames. A
        high value will classify more frames as noise but might remove desired
        signal!

    """

    def __init__(self, frame_len, lpc_order, iterations, alpha=0.8,
                 thresh=0.01):

        if frame_len % 2:
            raise ValueError("Frame length should be even as this method "
                             "relies on 50% overlap.")

        if (alpha > 1) or (alpha < 0):
            raise ValueError("`alpha` parameter should be within [0,1].")

        self.frame_len = frame_len
        self.hop = frame_len // 2
        self.lpc_order = lpc_order
        self.iterations = iterations
        self.alpha = alpha

        # simple energy-based voice activity detector
        self.thresh = thresh

        # initialize power spectral densities
        self.speech_psd = np.ones(self.hop+1)
        self.noise_psd = 0
        self.wiener_filt = np.ones(self.hop+1)

    def compute_filtered_output(self, current_frame, frame_dft=None):
        """
        Compute Wiener filter in the frequency domain.

        Parameters
        ----------
        current_frame : numpy array
            Noisy samples.
        frame_dft : numpy array
            DFT of input samples. If not provided, it will be computed.

        Returns
        -------
        numpy array
            Output of denoising in the frequency domain.
        """

        if frame_dft is None:
            frame_dft = np.fft.rfft(current_frame)
        frame_dft /= np.sqrt(self.frame_len)

        frame_energy = np.std(current_frame)**2

        # simple VAD
        if frame_energy < self.thresh:    # noise frame

            # update noise power spectral density
            # assuming white noise, i.e. flat spectrum
            self.noise_psd = self.alpha * self.noise_psd +\
                             (1 - self.alpha) * frame_energy

            # update wiener filter
            self.wiener_filt[:] = compute_wiener_filter(self.speech_psd,
                                                        self.noise_psd)
        else:   # speech frame

            s_i = current_frame

            # iteratively update speech power spectral density / wiener filter
            for i in range(self.iterations):
                a = lpc(s_i, self.lpc_order)
                g2 = compute_squared_gain(a, self.noise_psd, current_frame)
                self.speech_psd[:] = compute_speech_psd(a, g2, self.frame_len)

                # update Wiener filter
                self.wiener_filt[:] = compute_wiener_filter(self.speech_psd,
                                                            self.noise_psd)
                # update current frame with denoised version
                s_i = np.fft.irfft(self.wiener_filt * frame_dft)

        return self.wiener_filt * frame_dft


def compute_speech_psd(a, g2, nfft):
    """
    Compute power spectral density of speech as specified in equation (41b) of

        J. Lim and A. Oppenheim, *All-Pole Modeling of Degraded Speech,*
        IEEE Transactions on Acoustics, Speech, and Signal Processing 26.3
        (1978): 197-210.

    Namely:

    .. math::

        P_S(\\omega) = \dfrac{g^2}{\\left \| 1 - \sum_{k=1}^p a_k \cdot e^{-jk\omega} \\right \|^2},

    where :math:`p` is the LPC order, :math:`a_k` are the LPC coefficients, and
    :math:`g` is an estimated gain factor.

    The power spectral density is computed at the frequencies corresponding to
    a DFT of length `nfft`.

    Parameters
    ----------
    a : numpy array
        LPC coefficients.
    g2 : float
        Squared gain.
    nfft : int
        FFT length.

    Returns
    -------
    numpy array
        Power spectral density from LPC coefficients.
    """
    A = np.fft.rfft(np.r_[np.ones(1), -1*a], nfft)
    return g2 / np.abs(A)**2


def compute_squared_gain(a, noise_psd, y):
    """
    Estimate the squared gain of the speech power spectral density as done on
    p. 204 of

        J. Lim and A. Oppenheim, *All-Pole Modeling of Degraded Speech,*
        IEEE Transactions on Acoustics, Speech, and Signal Processing 26.3
        (1978): 197-210.

    Namely solving for :math:`g^2` such that the following expression is
    satisfied:

    .. math::

        \dfrac{N}{2\pi} \int_{-\pi}^{\pi} \dfrac{g^2}{\\left \| 1 - \sum_{k=1}^p a_k \cdot e^{-jk\omega} \\right \|^2} d\omega = \sum_{n=0}^{N-1} y^2(n) - N\cdot\sigma_d^2,


    where :math:`N` is the number of noisy samples :math:`y`, :math:`a_k`
    are the :math:`p` LPC coefficients, and :math:`\sigma_d^2` is the
    noise variance.

    Parameters
    ----------
    a : numpy array
        LPC coefficients.
    noise_psd : float or numpy array
        Noise variance if white noise, numpy array otherwise.
    y : numpy array
        Noisy time domain samples.

    Returns
    -------
    float
        Squared gain.
    """

    p = len(a)
    def _lpc_all_pole(omega):
        k = np.arange(p) + 1
        return 1 / np.abs(1 - np.dot(a, np.exp(-1j * k * omega))) ** 2

    N = len(y)

    # right hand side of expression
    if np.isscalar(noise_psd):   # white noise, i.e. flat spectrum
        rhs = np.sum(y**2) - N * noise_psd
    else:
        rhs = np.sum(y**2) - np.sum(noise_psd)

    # estimate integral
    d_omega = 2 * np.pi / 1000
    omega_vals = np.arange(-np.pi, np.pi, d_omega)
    vec_integrand = np.vectorize(_lpc_all_pole)
    integral = integrate.trapz(vec_integrand(omega_vals), omega_vals)
    return rhs * 2 * np.pi / N / integral


def compute_wiener_filter(speech_psd, noise_psd):
    """
    Compute Wiener filter in the frequency domain.

    Parameters
    ----------
    speech_psd : numpy array
        Speech power spectral density.
    noise_psd : float or numpy array
        Noise variance if white noise, numpy array otherwise.

    Returns
    -------
    numpy array
        Frequency domain filter, computed at the same frequency values as
        `speech_psd`.
    """

    return speech_psd / (speech_psd + noise_psd)


def apply_iterative_wiener(noisy_signal, frame_len=512, lpc_order=20,
                           iterations=2, alpha=0.8, thresh=0.01):
    """
    One-shot function to apply iterative Wiener filtering for denoising.

    Parameters
    ----------
    noisy_signal : numpy array
        Real signal in time domain.
    frame_len : int
        Frame length in samples. 50% overlap is used with hanning window.
    lpc_order : int
        Number of LPC coefficients to compute
    iterations : int
        How many iterations to perform in updating the Wiener filter for each
        signal frame.
    alpha : int
        Smoothing factor within [0,1] for updating noise level. Closer to `1`
        gives more weight to the previous noise level, while closer to `0`
        gives more weight to the current frame's level. Closer to `0` can track
        more rapid changes in the noise level. However, if a speech frame is
        incorrectly identified as noise, you can end up removing desired
        speech.
    thresh : float
        Threshold to distinguish between (signal+noise) and (noise) frames. A
        high value will classify more frames as noise but might remove desired
        signal!

    Returns
    -------
    numpy array
        Enhanced/denoised signal.
    """

    from pyroomacoustics import hann
    from pyroomacoustics.transform import STFT

    hop = frame_len // 2
    window = hann(frame_len, flag='asymmetric', length='full')
    stft = STFT(frame_len, hop=hop, analysis_window=window, streaming=True)
    scnr = IterativeWiener(frame_len, lpc_order, iterations, alpha, thresh)

    processed_audio = np.zeros(noisy_signal.shape)
    n = 0
    while noisy_signal.shape[0] - n >= hop:
        # SCNR in frequency domain
        stft.analysis(noisy_signal[n:(n + hop), ])
        X = scnr.compute_filtered_output(current_frame=stft.fft_in_buffer,
                                         frame_dft=stft.X)

        # back to time domain
        processed_audio[n:n + hop, ] = stft.synthesis(X)

        # update step
        n += hop

    return processed_audio





import numpy as np
from scipy.signal import fftconvolve

def trinicon(signals, w0=None,
        filter_length=2048,
        block_length=None, 
        n_blocks=8,
        alpha_on=None,
        j_max=10,
        delta_max=1e-4,
        sigma2_0=1e-7,
        mu=0.001,
        lambd_a=0.2,
        return_filters=False):
    '''
    Implementation of the TRINICON Blind Source Separation algorithm as described in

    R. Aichner, H. Buchner, F. Yan, and W. Kellermann  *A real-time
    blind source separation scheme and its application to reverberant and noisy
    acoustic environments*,  Signal Processing, 86(6), 1260-1277.
    doi:10.1016/j.sigpro.2005.06.022, 2006. `[pdf] <http://www.buchner-net.com/lnt2006_19.pdf>`_

    Specifically, adaptation of the pseudo-code from Table 1.

    The implementation is hard-coded for 2 output channels.

    Parameters
    ----------
    signals: ndarray (nchannels, nsamples)
        The microphone input signals (time domain)
    w0: ndarray (nchannels, nsources, nsamples), optional
        Optional initial value for the demixing filters
    filter_length: int, optional
        The length of the demixing filters, if w0 is provided, this option is ignored
    block_length: int, optional
        Block length (default 2x filter_length)
    n_blocks: int, optional
        Number of blocks processed at once (default 8)
    alpha_on: int, optional
        Online overlap factor (default ``n_blocks``)
    j_max: int, optional
        Number of offline iterations (default 10)
    delta_max: float, optional
        Regularization parameter, this sets the maximum value of the regularization term (default 1e-4)
    sigma2_0: float, optional
        Regularization parameter, this sets the reference (machine?) noise
        level in the regularization (default 1e-7)
    mu: float, optional
        Offline update step size (default 0.001)
    lambd_a: float, optional
        Online forgetting factor (default 0.2)
    return_filters: bool
        If true, the function will return the demixing matrix too (default False)

    Returns
    -------
    ndarray
        Returns an (nsources, nsamples) array. Also returns
        the demixing matrix (nchannels, nsources, nsamples)
        if ``return_filters`` keyword is True.
    '''

    P = signals.shape[0] # number of microphones
    Q = 2                # number of output channels


    # the filters
    if w0 is None:
        L = filter_length
        w = np.zeros((P,Q,L))
        w[:P//2,0,L//2] = 1
        w[P//2:,1,L//2] = 1
    else:
        w = w0.copy()
        L = w0.shape[2]

    K = n_blocks         # Number of successive blocks processed at the same time
    if block_length is None:  # Block length
        N = 2*L
    else:
        N = block_length

    if alpha_on is None:
        alpha_on = K         # online overlap factor

    hop = K * L // alpha_on

    # pad with zeros to have a whole number of online blocks
    if signals.shape[1] % hop != 0:
        signals = np.concatenate((signals, np.zeros((P, hop - (signals.shape[0]%hop)))), axis=1)

    S = signals.shape[1] # total signal length
    M = S / hop          # number of online blocks

    y = np.zeros((Q,S))    # the processed output signal

    m = 1               # online block index
    while m <= M:        # online loop
        
        # new chunk of input signal
        x = np.zeros((P,K*L+N))
        if m*hop > S:
            # we need some zero padding at the back
            le = S - (m-1)*hop + N
            x[:le] = signals[:,m*hop-K*L-N:]
        if m*hop >= K*L+N:
            x = signals[:,m*hop-K*L-N:m*hop]
        else:
            # we need some zero padding at the beginning
            x[:,-m*hop:] = signals[:,:m*hop]

        # use filter from previous iteration to initialize offline part
        w_new = w.copy()

        for j in range(j_max):     # offline update loop

            y_c = np.zeros((Q,K*L+N-L))  # c stands for chunk
            y_blocks = np.zeros((Q,K,N))

            for q in range(Q):
                # convolve with filters
                for p in range(P):
                    # We discard the 'oldest' output of the convolution according
                    # to the filter matrix definition (6) in the paper
                    y_c[q,:] += fftconvolve(x[p,:], w_new[p,q,:], mode='valid')[1:]

                # split into smaller blocks
                for i in range(K):
                    y_blocks[q,i,:] = y_c[q,i*L:i*L+N]

            # blocks energy
            sigma2 = np.sum(y_blocks**2, axis=2)

            # cross-correlations
            # XXX This is the part hard coded for two channels XXX
            r_cross = np.zeros((Q,K,2*L-1))
            for i in range(K):
                y0 = y_c[0,i*L:i*L+N]
                y1 = y_c[1,i*L:i*L+N]
                r = fftconvolve(y1, y0[::-1], mode='full')
                r_cross[0,i,:] = r[N-L:N+L-1]         # r_y1y0
                r_cross[1,i,:] = r_cross[0,i,::-1]    # r_y0y1 by symmetry is just r_y1y0 reversed

            # regularization term
            delta = delta_max*np.exp(-sigma2/sigma2_0)

            # offline update
            delta_w = np.zeros((P,Q,L))
            for q in range(Q):
                for p in range(P):
                    for i in range(K):
                        # this implements the row-wise sylvester constraint as explained in Fig. 4 (b) of paper
                        delta_w[p,q,:] += fftconvolve(r_cross[q,i,:]/(sigma2[q,i]+delta[q,i]), w_new[p,1-q,::-1], mode='valid')[::-1]
                    delta_w[p,q,:] /= K

            w_new = w_new - mu*delta_w

        # online update
        w = lambd_a*w + (1-lambd_a)*w_new

        # compute output signal
        for q in range(Q):
            for p in range(P):
                y[q,(m-1)*hop:m*hop] += fftconvolve(x[p,-hop-L+1:], w[p,q,:], mode='valid')
        
        # next block
        m += 1

    if return_filters:
        return y, w
    else:
        return y




import numpy as np
import pyroomacoustics as pra

room = pra.ShoeBox([4,6], fs=16000, max_order=1)

# add sources in the room
room.add_source([2, 1.5])  # nice source
room.add_source([2,4.5])   # interferer

# add a circular beamforming array
shape = pra.circular_2D_array([2.5,3], 8, 0., 0.15)
bf = pra.Beamformer(shape, room.fs, Lg=500)
room.add_microphone_array(bf)

# run the ISM
room.image_source_model()

# the noise matrix, note that the size is the number of
# sensors multiplied by the filter size
Rn = np.eye(bf.M * bf.Lg) * 1e-5

def test_rake_max_udr_filters():
    # no interferer
    bf.rake_max_udr_filters(room.sources[0][:4], R_n=Rn, delay=0.015, epsilon=1e-2)
    # with interferer
    bf.rake_max_udr_filters(room.sources[0][:4], interferer=room.sources[1][:4], R_n=Rn, delay=0.015, epsilon=1e-2)

def test_perceptual_filters():
    # no interferer
    bf.rake_perceptual_filters(room.sources[0][:4], R_n=Rn)
    # with interferer
    bf.rake_perceptual_filters(room.sources[0][:4], interferer=room.sources[1][:4], R_n=Rn)
    pass

if __name__ == '__main__':

    import matplotlib.pyplot as plt

    bf.rake_perceptual_filters(room.sources[0][:4], interferer=room.sources[1][:4], R_n=Rn, epsilon=0.1)
    bf.plot()

    room.plot(img_order=1, freq=[700., 1000., 2000.])

    plt.show()



'''
This module contains useful functions to compute distances and errors on on
circles and spheres.
'''
from __future__ import division
import numpy as np


def circ_dist(azimuth1, azimuth2, r=1.):
    ''' 
    Returns the shortest distance between two points on a circle

    Parameters
    ----------
    azimuth1:
        azimuth of point 1
    azimuth2:
        azimuth of point 2
    r: optional
        radius of the circle (Default 1)
    '''
    return np.arccos(np.cos(azimuth1 - azimuth2))


def great_circ_dist(r, colatitude1, azimuth1, colatitude2, azimuth2):
    """
    calculate great circle distance for points located on a sphere

    Parameters
    ----------
    r: radius of the sphere
    colatitude1: colatitude of point 1
    azimuth1: azimuth of point 1
    colatitude2: colatitude of point 2
    azimuth2: azimuth of point 2
    
    Returns 
    -------
    float or ndarray
        great-circle distance
    """
    d_azimuth = np.abs(azimuth1 - azimuth2)
    dist = r * np.arctan2(np.sqrt((np.sin(colatitude2) * np.sin(d_azimuth)) ** 2 +
                                  (np.sin(colatitude1) * np.cos(colatitude2) -
                                   np.cos(colatitude1) * np.sin(colatitude2) * np.cos(d_azimuth)) ** 2),
                          np.cos(colatitude1) * np.cos(colatitude2) +
                          np.sin(colatitude1) * np.sin(colatitude2) * np.cos(d_azimuth))
    return dist

def spher2cart(r, azimuth, colatitude):
    """
    Convert a spherical point to cartesian coordinates.

    Parameters
    ----------
    r:
        radius
    azimuth:
        azimuth
    colatitude:
        colatitude

    Returns
    -------
    ndarray
        An ndarray containing the Cartesian coordinates of the points its columns
    """
    # convert to cartesian
    x = r * np.cos(azimuth) * np.sin(colatitude)
    y = r * np.sin(azimuth) * np.sin(colatitude)
    z = r * np.cos(colatitude)
    return np.array([x, y, z])


def polar_distance(x1, x2):
    """
    Given two arrays of numbers x1 and x2, pairs the cells that are the
    closest and provides the pairing matrix index: x1(index(1,:)) should be as
    close as possible to x2(index(2,:)). The function outputs the average of 
    the absolute value of the differences abs(x1(index(1,:))-x2(index(2,:))).

    Parameters
    ----------
    x1: 
        vector 1
    x2: 
        vector 2

    Returns
    -------
    d:
        minimum distance between d
    index:
        the permutation matrix
    """
    x1 = np.reshape(x1, (1, -1), order='F')
    x2 = np.reshape(x2, (1, -1), order='F')
    N1 = x1.size
    N2 = x2.size
    diffmat = np.arccos(np.cos(x1 - np.reshape(x2, (-1, 1), order='F')))
    min_N1_N2 = np.min([N1, N2])
    index = np.zeros((min_N1_N2, 2), dtype=int)
    if min_N1_N2 > 1:
        for k in range(min_N1_N2):
            d2 = np.min(diffmat, axis=0)
            index2 = np.argmin(diffmat, axis=0)
            index1 = np.argmin(d2)
            index2 = index2[index1]
            index[k, :] = [index1, index2]
            diffmat[index2, :] = float('inf')
            diffmat[:, index1] = float('inf')
        d = np.mean(np.arccos(np.cos(x1[:, index[:, 0]] - x2[:, index[:, 1]])))
    else:
        d = np.min(diffmat)
        index = np.argmin(diffmat)
        if N1 == 1:
            index = np.array([1, index])
        else:
            index = np.array([index, 1])
    return d, index



import shutil
from pyroomacoustics.datasets.utils import download_uncompress
import ssl

test_url = 'https://github.com/LCAV/pyroomacoustics/archive/master.tar.gz'
extracted_name = 'pyroomacoustics-master'


def test_download_uncompress():
    context = ssl._create_unverified_context()
    download_uncompress(test_url, context=context)
    shutil.rmtree(extracted_name, ignore_errors=True)


if __name__ == '__main__':
    test_download_uncompress()




import numpy as np
import pyroomacoustics as pra

def test_room_is_inside():
    # fix the seed for repeatable testing
    np.random.seed(0)

    # This was a problematic case
    # if the source is placed at the same height as one of the corners
    # the test would fail, even though the source is in the room

    floorplan = [ [0, 6, 6, 2, 0],
                  [0, 0, 5, 5, 3] ]

    source_loc = [ 2, 3 ] # same y-coordinate as the corner at [0, 3]

    room = pra.Room.from_corners(floorplan)
    room.add_source(source_loc)

    for i in range(100):
        # because the test is randomized, let's check many times

        assert room.is_inside([0,0], include_borders=True)
        assert not room.is_inside([0,0], include_borders=False)

        assert room.is_inside([3,0], include_borders=True)
        assert not room.is_inside([3,0], include_borders=False)

        assert room.is_inside([1,4], include_borders=True)
        assert not room.is_inside([1,4], include_borders=False)

        assert room.is_inside([0,1], include_borders=True)
        assert not room.is_inside([0,1], include_borders=False)

        assert not room.is_inside([0.5,4], include_borders=False)

    # now test in 3D
    room.extrude(4.)

    for i in range(100):
        # because the test is randomized, let's check many times

        assert room.is_inside([2, 3, 1.7])

        assert not room.is_inside([0.5, 4, 1.8])

        assert not room.is_inside([0.5, 4, 1.8])

        assert room.is_inside([0,0,0], include_borders=True)
        assert not room.is_inside([0,0,0], include_borders=False)

        assert room.is_inside([3,0,0], include_borders=True)
        assert not room.is_inside([3,0,0], include_borders=False)

        assert room.is_inside([0,1,0], include_borders=True)
        assert not room.is_inside([0,1,0], include_borders=False)

        assert room.is_inside([3,2,0], include_borders=True)
        assert not room.is_inside([3,2,0], include_borders=False)

        assert room.is_inside([1,4,3], include_borders=True)
        assert not room.is_inside([1,4,3], include_borders=False)

        assert not room.is_inside([2,2,7])
        assert not room.is_inside([2,2,-7])

if __name__ == '__main__':

    test_room_is_inside()



# Copyright (c) 2019 Kouhei Sekiguchi, Yoshiaki Bando
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
FastMNMF
========

Blind Source Separation based on Fast Multichannel Nonnegative Matrix Factorization (FastMNMF)
"""
import numpy as np


def fastmnmf(
    X,
    n_src=None,
    n_iter=30,
    W0=None,
    n_components=4,
    callback=None,
    mic_index=0,
    interval_update_Q=1,
    interval_normalize=10,
    initialize_ilrma=False,
):
    """
    Implementation of FastMNMF algorithm presented in

    K. Sekiguchi, A. A. Nugraha, Y. Bando, K. Yoshii, *Fast Multichannel Source 
    Separation Based on Jointly Diagonalizable Spatial Covariance Matrices*, EUSIPCO, 2019. [`arXiv <https://arxiv.org/abs/1903.03237>`_]

    The code of FastMNMF with GPU support and FastMNMF-DP which integrates DNN-based source model
    into FastMNMF is available on https://github.com/sekiguchi92/SpeechEnhancement 

    Parameters
    ----------
    X: ndarray (nframes, nfrequencies, nchannels)
        STFT representation of the observed signal 
    n_src: int, optional
        The number of sound sources (if n_src = None, n_src is set to the number of microphone)
    n_iter: int, optional
        The number of iterations
    W0: ndarray (nfrequencies, nchannels, nchannels), optional
        Initial value for diagonalizer Q
        Demixing matrix can be used as the initial value
    n_components: int, optional
        Number of components in the non-negative spectrum
    callback: func, optional
        A callback function called every 10 iterations, allows to monitor convergence
    mic_index: int, optional
        The index of microphone of which you want to get the source image
    interval_update_Q: int, optional
        The interval of updating Q
    interval_normalize: int, optional
        The interval of parameter normalization
    initialize_ilrma: boolean, optional
        Initialize diagonalizer Q by using ILRMA

    Returns
    -------
    separated spectrogram: numpy.ndarray
        An (nframes, nfrequencies, nsources) array. 
    """

    eps = 1e-7

    # initialize parameter
    X_FTM = X.transpose(1, 0, 2)
    n_freq, n_frames, n_chan = X_FTM.shape
    XX_FTMM = np.matmul(X_FTM[:, :, :, None], X_FTM[:, :, None, :].conj())
    if n_src is None:
        n_src = X_FTM.shape[
            2
        ]  # determined case (the number of source = the number of microphone)

    if initialize_ilrma:  # initialize by using ILRMA
        from pyroomacoustics.bss.ilrma import ilrma

        Y_TFM, W = ilrma(
            X, n_iter=10, n_components=2, proj_back=False, return_filters=True
        )
        Q_FMM = W
        sep_power_M = np.abs(Y_TFM).mean(axis=(0, 1))
        g_NFM = np.ones([n_src, n_freq, n_chan]) * 1e-2
        for n in range(n_src):
            g_NFM[n, :, sep_power_M.argmax()] = 1
            sep_power_M[sep_power_M.argmax()] = 0
    elif W0 != None:  # initialize by W0
        Q_FMM = W0
        g_NFM = np.ones([n_src, n_freq, n_chan]) * 1e-2
        for m in range(n_chan):
            g_NFM[m % n_src, :, m] = 1
    else:  # initialize by using observed signals
        Q_FMM = np.tile(np.eye(n_chan).astype(np.complex), [n_freq, 1, 1])
        g_NFM = np.ones([n_src, n_freq, n_chan]) * 1e-2
        for m in range(n_chan):
            g_NFM[m % n_src, :, m] = 1
    for m in range(n_chan):
        mu_F = (Q_FMM[:, m] * Q_FMM[:, m].conj()).sum(axis=1).real
        Q_FMM[:, m] = Q_FMM[:, m] / np.sqrt(mu_F[:, None])
    H_NKT = np.random.rand(n_src, n_components, n_frames)
    W_NFK = np.random.rand(n_src, n_freq, n_components)
    lambda_NFT = np.matmul(W_NFK, H_NKT)
    Qx_power_FTM = (
        np.abs((np.matmul(Q_FMM[:, None], X_FTM[:, :, :, None]))[:, :, :, 0]) ** 2
    )
    Y_FTM = (lambda_NFT[..., None] * g_NFM[:, :, None]).sum(axis=0)

    separated_spec = np.zeros([n_src, n_freq, n_frames], dtype=np.complex)

    def separate():
        Qx_FTM = (Q_FMM[:, None] * X_FTM[:, :, None]).sum(axis=3)
        diagonalizer_inv_FMM = np.linalg.inv(Q_FMM)
        tmp_NFTM = lambda_NFT[..., None] * g_NFM[:, :, None]
        for n in range(n_src):
            tmp = (
                np.matmul(
                    diagonalizer_inv_FMM[:, None],
                    (Qx_FTM * (tmp_NFTM[n] / (tmp_NFTM).sum(axis=0)))[..., None],
                )
            )[:, :, mic_index, 0]
            separated_spec[n] = tmp
        return separated_spec.transpose(2, 1, 0)

    # update parameters
    for epoch in range(n_iter):
        if callback is not None and epoch % 10 == 0:
            callback(separate())

        # update_WH (basis and activation of NMF)
        tmp_yb1 = (g_NFM[:, :, None] * (Qx_power_FTM / (Y_FTM ** 2))[None]).sum(
            axis=3
        )  # [N, F, T]
        tmp_yb2 = (g_NFM[:, :, None] / Y_FTM[None]).sum(axis=3)  # [N, F, T]
        a_1 = (H_NKT[:, None, :, :] * tmp_yb1[:, :, None]).sum(axis=3)  # [N, F, K]
        b_1 = (H_NKT[:, None, :, :] * tmp_yb2[:, :, None]).sum(axis=3)  # [N, F, K]
        W_NFK *= np.sqrt(a_1 / b_1)

        a_1 = (W_NFK[:, :, :, None] * tmp_yb1[:, :, None]).sum(axis=1)  # [N, K, T]
        b_1 = (W_NFK[:, :, :, None] * tmp_yb2[:, :, None]).sum(axis=1)  # [N, F, K]
        H_NKT *= np.sqrt(a_1 / b_1)

        np.matmul(W_NFK, H_NKT, out=lambda_NFT)
        np.maximum(lambda_NFT, eps, out=lambda_NFT)
        Y_FTM = (lambda_NFT[..., None] * g_NFM[:, :, None]).sum(axis=0)

        # update diagonal element of spatial covariance matrix
        a_1 = (lambda_NFT[..., None] * (Qx_power_FTM / (Y_FTM ** 2))[None]).sum(
            axis=2
        )  # N F T M
        b_1 = (lambda_NFT[..., None] / Y_FTM[None]).sum(axis=2)
        g_NFM *= np.sqrt(a_1 / b_1)
        Y_FTM = (lambda_NFT[..., None] * g_NFM[:, :, None]).sum(axis=0)

        # udpate Diagonalizer which jointly diagonalize spatial covariance matrix
        if (interval_update_Q <= 0) or ((epoch + 1) % interval_update_Q == 0):
            for m in range(n_chan):
                V_FMM = (XX_FTMM / Y_FTM[:, :, m, None, None]).mean(axis=1)
                tmp_FM = np.linalg.solve(
                    np.matmul(Q_FMM, V_FMM), np.eye(n_chan)[None, m]
                )
                Q_FMM[:, m] = (
                    tmp_FM
                    / np.sqrt(
                        ((tmp_FM.conj()[:, :, None] * V_FMM).sum(axis=1) * tmp_FM).sum(
                            axis=1
                        )
                    )[:, None]
                ).conj()
            Qx_power_FTM = (
                np.abs((np.matmul(Q_FMM[:, None], X_FTM[:, :, :, None]))[:, :, :, 0])
                ** 2
            )

        # normalize
        if (interval_normalize <= 0) or (epoch % interval_normalize == 0):
            phi_F = np.sum(Q_FMM * Q_FMM.conj(), axis=(1, 2)).real / n_chan
            Q_FMM /= np.sqrt(phi_F)[:, None, None]
            g_NFM /= phi_F[None, :, None]

            mu_NF = (g_NFM).sum(axis=2)
            g_NFM /= mu_NF[:, :, None]
            W_NFK *= mu_NF[:, :, None]

            mu_NK = W_NFK.sum(axis=1)
            W_NFK /= mu_NK[:, None]
            H_NKT *= mu_NK[:, :, None]
            np.matmul(W_NFK, H_NKT, out=lambda_NFT)
            np.maximum(lambda_NFT, 1e-10, out=lambda_NFT)

            Qx_power_FTM = (
                np.abs((np.matmul(Q_FMM[:, None], X_FTM[:, :, :, None]))[:, :, :, 0])
                ** 2
            )
            Y_FTM = (lambda_NFT[..., None] * g_NFM[:, :, None]).sum(axis=0)

    return separate()
